/*------------------------------------
 * WinHotkeys.c -  Fekete Krisztian
 * fekete.krisztian@freemail.hu
 *------------------------------------
 *
 * HISTORY:
 * 2001/06/28: unix X version
 * 2000/07/27: fixed regexp bug: last used window is brought up if the window is not found
 * 2000/07/26: added regexp support, startup directory support
 * 2000/07/06: removed msgbox when unregistering hotkeys
 * 2000/07/05: added init file support
 * compile with
 *        gcc -mno-cygwin -mwindows WinHotkeys.c str_lib.c hotkey.c -o WinHotkeys.exe
 */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "regex.h"

#include "str_lib.h"
#include "hotkey.h"
#include "xlib.h"
#include "os.h"

def register_hotkey(
     Hotkey* hk)
{
     if (hk) {
          if (gui_register_hotkey(hk)) {
               IFDEBUG(printf("register_hotkey: <%s> succeeded", hk->hk_name));
               succeed;
          } else {
               gui_message("register hotkey FAILED", hk->hk_name);
               hotkey_free(&hk);
               fail;
          }
     }
     
     fail;
}

static Hotkey*
make_hotkey_copy(
     Hotkey* hk)
{
     Hotkey* new_hk;
    
     new_hk = (Hotkey *)malloc(sizeof(Hotkey));
     if (new_hk) {
          memcpy(new_hk, hk, sizeof(Hotkey));
     }
    
     return new_hk;
}

def load_hotkeys(
     const char *fname)
{
     FILE* f;
     Hotkey hk;
     yesno ok;
     
     IFDEBUG(printf("load_hotkeys: %s", fname));
     f = fopen(fname, "r");
     if (!f) {
          gui_message("Cannot open file", fname);
          fail;
     }
     while (ok && hotkey_read(f, &hk)) {
          ok = ok && register_hotkey(make_hotkey_copy(&hk));
     }
     ok = ok && feof(f);
     fclose(f);
  
     return ok;
}

def regexp_match(
     const char* pattern,
     const char* string)
{
     yesno found;
     int error;
     regex_t re;
     regex_t* r = &re;
     
     error = regcomp(r, pattern, REG_EXTENDED | REG_NOSUB);
     if (error) {
          fail;
     } else {
          int n = r->re_nsub + 1;
          regmatch_t* m = (regmatch_t*) malloc(sizeof(regmatch_t) * n);
          found = REG_NOMATCH != regexec(r, string, n, m, 0);
          free(m);
     }
     regfree(r);
     
     return found;
}

def match(
     const Pattern* pattern,
     const char* str)
{
     if (0 == pattern->pattern) {
          succeed;
     }
     switch (pattern->type) {
     case PT_REGEXP:
          return regexp_match(pattern->pattern, str);
          
     case PT_EXACT:
     default:
          if (streq(pattern->pattern, str)) {
               succeed;
          } else {
               fail;
          }
     }
}

def win_match(
     const Hotkey* hk,
     GUI_window w)
{
     char* str;
     yesno ok;
     
     ok = yes;
     IFDEBUG(printf("win_match: %s", hk->hk_name));
     if (ok && hk->winclass.pattern) {
          ok = ok && gui_get_window_class(w, &str);
          if (ok) {
               ok = ok && match(&(hk->winclass), str);
               free(str);
          }
     }
     if (ok && hk->winname.pattern) {
          ok = ok && gui_get_window_name(w, &str);
          if (ok) {
               ok = ok && match(&(hk->winname), str);
               free(str);
          }
     }
     if (ok && hk->wintitle.pattern) {
          ok = ok && gui_get_window_title(w, &str);
          if (ok) {
               ok = ok && match(&(hk->wintitle), str);
               free(str);
          }
     }
     if (ok && hk->wincommand.pattern) {
          ok = ok && gui_get_window_command(w, &str);
          if (ok) {
               ok = ok && match(&(hk->wincommand), str);
               free(str);
          }
     }
     
     return ok;
}

def find_window(
     const Hotkey* hk,
     GUI_window* w)
{
     GUI_window* window_list;
     int n;
     
     if (gui_get_window_list(&window_list, &n)) {
          int i;
          for (i = 0; i < n; i ++) {
               if (win_match(hk, window_list[i])) {
                    *w = window_list[i];
                    free(window_list);
                    succeed;
               }
          }
          free(window_list);
     }
     fail;
}

def iconify_top_window()
{
     GUI_window* window_list;
     GUI_window w;
     int n;
     yesno ok;
     
     ok = no;
     if (gui_get_window_list(&window_list, &n)) {
          while (n > 0) {
               w = window_list[n-1];
               if (gui_is_visible(w)) {
                    IFDEBUG(printf("VISIBLE: %ld", w));
                    ok = gui_iconify(w);
                    break;
               }
               n --;
               IFDEBUG(printf("NOT VISIBLE: %ld", w));
          }
          free(window_list);
     }
     return ok;
}

def exec_hotkey(
     Hotkey* hk)
{
     GUI_window w;
     
     IFDEBUG(printf("exec_hotkey: %s", hk->hk_name));
     os_reap_zombies();
     if (hotkey_has_pattern(hk) && find_window(hk, &w)) {
          return gui_bring_to_front(w);
     } else if (hk->iconify) {
          return iconify_top_window();
     } else if (hk->prog) {
          return os_spawn_with_dir(hk->dir, hk->prog);
     } else {
          fail;
     }
}

const char* prog_name = "X11Hotkeys";
yesno debug = no;

def usage()
{
     gui_message(
          "\nUsage",
          "\n"
          "X11Hotkeys [options]\n"
          "	options are:\n"
          "		-d        to turn on debugging output - you can catch errors in the input file\n"
          "		-f <file> read in hotkey definitions from \"file\", you can have only one -f\n");
     exit(3);
     fail;
}

int 
main(
     int argc,
     char** argv)
{
     char* fname;
     int i;
     
     if (!gui_init()) {
          gui_message("ERROR", "Can not open X display");
          return 1;
     }
     
     /* process arguments */
     fname = 0;
     IFDEBUG(printf("argc: %d", argc));
     for (i = 1; i < argc; i ++) {
          if (streq(argv[i], "-d")) {
               debug = yes;
               IFDEBUG(printf("DEBUGGING ON!"));
          } else if (streq(argv[i], "-f")) {
               if (fname) {
                    usage();
               }
               i ++;
               if (i < argc) {
                    fname = argv[i];
               } else {
                    gui_message("ERROR", "no argument for -f");
                    usage();
               }
          } else {
               gui_message("unrecognized option", argv[i]);
               usage();
          }
     }
     
     /* if no filename was given try .xhotkeys in the user's home */
     if (0 == fname) {
          const char* xhotkeys = ".xhotkeys";
          char* home;
          
          IFDEBUG(printf("no hotkey definition specified, trying default in user's home"));
          home = getenv("HOME");
          if (0 == home) {
               gui_message("ERROR", "HOME not set!");
               exit(4);
          }
          fname = malloc(strlen(home) + 1 + strlen(xhotkeys) + 1);
          if (fname) {
               sprintf(fname, "%s/%s", home, xhotkeys);
               if (0 != access(fname, R_OK)) {
                    gui_message("cannot read", fname);
                    exit(5);
               }
          }
     }
     
     /* load & register hotkeys */
     if (fname) {
          if (!load_hotkeys(fname)) {
               gui_message("Can not load hotkey definition", fname);
               gui_unregister_hotkeys();
               gui_uninit();
               exit(6);
          }
     } else {
          usage();
     }
     
     IFDEBUG(printf("hotkeys were loaded successfully, starting event loop\n\n"));
     gui_run_event_loop(exec_hotkey);
     gui_unregister_hotkeys();
     gui_uninit();
     return 0;
}

/* end of file */
