/*------------------------------------
 * hotkey.c -  Fekete Krisztian
 * fekete.krisztian@freemail.hu
 *------------------------------------
 *
 *
 * HISTORY:
 * 2001/07/01: added support for wm_command (X11)
 * 2001/06/24: reworked for clarity and generality
 * 2000/07/27: added support for more keys
 * 2000/07/26: added dir and regexp support
 * 2000/07/05: initial version
 */

/* { pine - Mail
 *   prog: xterm -e pine
 *   title: pine
 *   class: XTerm
 *   key: control alt m
 * }
 */

#include <string.h>
#include <stdlib.h>
#include <string.h>
#include "hotkey.h"
#include "str_lib.h"

typedef enum {
     HK_COMMENT,
     HK_BEGIN,
     HK_HOTKEY,

     HK_PT_EXACT,
     HK_PT_REGEXP,
     
     HK_WINNAME,
     HK_WINCLASS,
     HK_WINTITLE,
     HK_WINCOMMAND,
     
     HK_ICONIFY,
     HK_PROG,
     HK_DIR,
     
     HK_END
} Hotkey_symbol;

typedef struct {
     const char* name;
     Hotkey_symbol symbol;
} String_to_hotkey_sym;

static String_to_hotkey_sym string_to_hotkey_sym_map[] = {
     {"{",         HK_BEGIN},
     
     {"key:",      HK_HOTKEY},
     
     {"pattern-type-exact",  HK_PT_EXACT},
     {"pattern-type-regexp", HK_PT_REGEXP},
     /* {"pattern-type-substring", HK_PT_SUBSTRING}, */
     
     {"name:",     HK_WINNAME},
     {"class:",    HK_WINCLASS},
     {"title:",    HK_WINTITLE},
     {"command:",  HK_WINCOMMAND},
     
     {"prog:",     HK_PROG},
     {"spawn:",    HK_PROG},
     {"dir:",      HK_DIR},
     
     {"iconify!",  HK_ICONIFY},
     
     {"}",         HK_END},
     {0, 0}
};

decl cleanup_rd_hk(Hotkey* hk, string* line);
decl hotkey_init(Hotkey* hk);
decl string_to_hotkey_sym(string* word, Hotkey_symbol* sym);

def string_to_hotkey_sym(
     string* word,
     Hotkey_symbol* sym)
{
     char* str;
     String_to_hotkey_sym* p;
     
     str_to_cstr(word, &str);
     p = string_to_hotkey_sym_map;
     while (p->name) {
          if (0 == strcmp(p->name, str)) {
               IFDEBUG(printf("cmd: %s", str));
               *sym = p->symbol;
               succeed;
          }
          p++;
     }
     *sym = HK_COMMENT;
     
     succeed;
}

def pattern_init(
     Pattern* p)
{
     p->pattern = 0;
     p->type = PT_EXACT;
     succeed;
}

def hotkey_init(
     Hotkey* hk)
{
     hk->key_code = 0;
     hk->modifiers = 0;
     pattern_init(&(hk->wintitle));
     pattern_init(&(hk->winclass));
     pattern_init(&(hk->winname));
     pattern_init(&(hk->wincommand));
     hk->iconify = no;
     hk->prog = 0;
     hk->dir = 0;
     hk->hk_name = 0;
     succeed;
}

def hk_set_str(
     char** dest,
     string* str)
{
     char* cstr;
     
     str_to_cstr(str, &cstr);
     
     if (*dest) {
          free(*dest);
     }
     *dest = cstr;
     
     succeed;
}

def hk_set_pattern(
     Pattern* pattern,
     string* str,
     Pattern_type pt)
{
     hk_set_str(&(pattern->pattern), str);
     pattern->type = pt;
     succeed;
}

typedef struct {
     const char* str;
     GUI_modifier mod;
} String_to_modifier;

static String_to_modifier string_to_modifier_map[] = {
     {"shift",   GUI_SHIFT},
     {"control", GUI_CONTROL},
     {"ctrl",    GUI_CONTROL},
     {"alt",     GUI_ALT},
     {"meta",    GUI_META},
     {"super",   GUI_SUPER},
     {"hyper",   GUI_HYPER},
     {0, 0}
};

def string_to_modifier(
     char* str,
     GUI_modifier_mask* mask)
{
     String_to_modifier* p;
     
     p = string_to_modifier_map;
     while (p->str) {
          if (0 == strcmp(p->str, str)) {
               *mask |= p->mod;
               IFDEBUG(printf("mod: %s", str));
               succeed;
          }
          p++;
     }
     fail;
}

def hotkey_read_key_seq(
     Hotkey* hk,
     string* line)
{
     string word;
     char* str;
     hk->key_code = 0;
     hk->modifiers = 0;
     
     for (;;) {
          if (!read_word(line, &word)) {
               fail;
          }
          str_to_cstr(&word, &str);
          IFDEBUG(printf("key: %s", str));
          if (!string_to_modifier(str, &(hk->modifiers))) {
               if (!gui_string_to_key(str, &(hk->key_code))) {
                    if (0 == strlen(str)) {
                         free(str);
                         succeed;
                    } else {
                         free(str);
                         fail;
                    }
               }
          }
          free(str);
     }
}

def free_str(
     char** str)
{
     if (*str) {
          free(*str);
     }
     *str = 0;
     succeed;
}

def free_pattern(
     Pattern* pattern)
{
     free_str(&(pattern->pattern));
     pattern->type = PT_EXACT;
     succeed;
}

defglob hotkey_free(
     Hotkey** hk)
{
     IFDEBUG(printf("hotkey_free: %p: %s", (void*)*hk, (*hk)->hk_name));
     if (*hk) {
          free_str(&((*hk)->hk_name));
          free_str(&((*hk)->prog));
          free_str(&((*hk)->dir));
          free_pattern(&((*hk)->winclass));
          free_pattern(&((*hk)->winname));
          free_pattern(&((*hk)->wintitle));
          free_pattern(&((*hk)->wincommand));
          free(*hk);
     }
     (*hk) = 0;
     succeed;
}

defglob hotkey_has_action(
     const Hotkey* hk)
{
     if (hk->iconify || hk->prog || hotkey_has_pattern(hk)) {
          succeed;
     } else {
          fail;
     }
}

defglob hotkey_has_pattern(
     const Hotkey* hk)
{
     if (hk->winclass.pattern
         || hk->wintitle.pattern
         || hk->winname.pattern
         || hk->wincommand.pattern) {
          succeed;
     } else {
          fail;
     }
}

def cleanup_rd_hk(
     Hotkey* hk,
     string* line)
{
     if (line) {
          str_uninit(line);
     }
     if (hk) {
          free_str(&(hk->prog));
          free_str(&(hk->dir));
          free_pattern(&(hk->winclass));
          free_pattern(&(hk->winname));
          free_pattern(&(hk->wintitle));
          free_pattern(&(hk->wincommand));
     }
    
     succeed;
}

defglob hotkey_read(
     FILE* f,
     Hotkey* hk)
{
     Hotkey_symbol sym;
     string line;
     string word;
     yesno processing;
     Pattern_type pt;
     
     pt = PT_EXACT;
     hotkey_init(hk);
     for (processing = no; !feof(f);) {
          /* read the next line of input */
          if (!read_line(f, &line)) {
               fail;
          }
          if (!read_word(&line, &word)) {
               cleanup_rd_hk(hk, &line);
               fail;
          }
          
          string_to_hotkey_sym(&word, &sym);
          switch (sym) {
          case HK_COMMENT:
               break;
          case HK_BEGIN:
               processing = yes;
               hk_set_str(&(hk->hk_name), &line);
               break;
          case HK_HOTKEY:
               hotkey_read_key_seq(hk, &line);
               break;
               
          case HK_PT_EXACT:
               pt = PT_EXACT;
               break;
          case HK_PT_REGEXP:
               pt = PT_REGEXP;
               break;
               
          case HK_WINNAME:
               hk_set_pattern(&(hk->winname), &line, pt);
               break;
          case HK_WINCLASS:
               hk_set_pattern(&(hk->winclass), &line, pt);
               break;
          case HK_WINTITLE:
               hk_set_pattern(&(hk->wintitle), &line, pt);
               break;
          case HK_WINCOMMAND:
               hk_set_pattern(&(hk->wincommand), &line, pt);
               break;
               
          case HK_PROG:
               hk_set_str(&(hk->prog), &line);
               break;
          case HK_DIR:
               hk_set_str(&(hk->dir), &line);
               break;
          case HK_ICONIFY:
               hk->iconify = yes;
               break;
               
          case HK_END:
               /* if the hk was read ok, the we succeed, otherwise clear hk and fail */
               if (hk->key_code && hk->modifiers && hotkey_has_action(hk)) {
                    IFDEBUG(printf("hotkey_read: SUCCEED"));
                    cleanup_rd_hk(0, &line);
                    succeed;
               } else {
                    IFDEBUG(printf("hotkey_read: FAIL"));
                    cleanup_rd_hk(hk, &line);
                    fail;
               }
               break;
          default:
               assert(0);
          }
          cleanup_rd_hk(0, &line);
     }
     fail;
}

/* end of file */
