/*------------------------------------
 * xlib.c -  Fekete Krisztian
 * fekete.krisztian@freemail.hu
 *------------------------------------
 */

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>             /* malloc(), free() */
#include <string.h>             /* str*() */
#include <unistd.h>
extern char* strdup(const char* str);

#include "xlib.h"

decl gui_get_class_hint(GUI_window w, XClassHint** class_hint);
decl gui_free_class_hint(XClassHint** class_hint);
decl gui_init_modifier_map();

static Display* gui_display;
static int gui_screen;
static Window gui_root;
yesno gui_error_occured = no;
static int gui_error_code;
static Hotkey** hotkeys = 0;
static int nhotkeys = 0;
static int hotkeys_size = 0;

typedef struct
{
  unsigned int meta_mask;
  unsigned int alt_mask;
  unsigned int super_mask;
  unsigned int hyper_mask;
} GUI_modifier_info;
GUI_modifier_info modifier_info;

defglob gui_clear_error()
{
     gui_error_occured = no;
     succeed;
}

defglob gui_get_error_text(
     char* buffer,
     int length)
{
     XGetErrorText(gui_display, gui_error_code, buffer, length);
     succeed;
}

def gui_default_error_handler(
     Display* display,
     XErrorEvent* ev)
{
     IFDEBUG(printf("X ERROR!!!"));
     gui_error_code = ev->error_code;
     gui_error_occured = yes;
     fail;
}

defglob gui_init()
{
     /* open connection with the X server. */
     gui_display = XOpenDisplay(0);
     if (0 == gui_display) {
          fail;
     }
     
     /*XErrorHandler*/
     XSetErrorHandler(gui_default_error_handler);
     
     gui_screen = DefaultScreen(gui_display);
     gui_root = RootWindow(gui_display, gui_screen);
     gui_init_modifier_map();
     
     succeed;
}

defglob gui_uninit()
{
     XCloseDisplay(gui_display);
     gui_display = 0;
     gui_screen = 0;
     gui_root = 0;
     succeed;
}

defglob gui_string_to_key(
     const char* str,
     GUI_key_code* key_code)
{
     KeySym key_sym;
     
     key_sym = XStringToKeysym(str);
     if (NoSymbol == key_sym) {
          fail;
     }
     
     *key_code = XKeysymToKeycode(gui_display, key_sym);
     succeed;
}

def gui_modmask_to_xmodmask(
     GUI_modifier_mask modifiers,
     unsigned int* x_mod_mask)
{
     *x_mod_mask = 0;
     if (modifiers & GUI_SHIFT) {
          *x_mod_mask |= ShiftMask;
     }
     if (modifiers & GUI_CONTROL) {
          *x_mod_mask |= ControlMask;
     }
     if (modifiers & GUI_ALT) {
          *x_mod_mask |= modifier_info.alt_mask;
          IFDEBUG(printf("alt_mask: %d", modifier_info.alt_mask));
     }
     if (modifiers & GUI_META) {
          *x_mod_mask |= modifier_info.meta_mask;
     }
     if (modifiers & GUI_SUPER) {
          *x_mod_mask |= modifier_info.super_mask;
     }
     if (modifiers & GUI_HYPER) {
          *x_mod_mask |= modifier_info.hyper_mask;
     }
     succeed;
}

def gui_xmodmask_to_modmask(
     unsigned int x_modifiers,
     GUI_modifier_mask* modifiers)
{
     *modifiers = 0;
     if (x_modifiers & ShiftMask) {
          *modifiers |= GUI_SHIFT;
     }
     if (x_modifiers & ControlMask) {
          *modifiers |= GUI_CONTROL;
     }
     if (x_modifiers & modifier_info.alt_mask) {
          *modifiers |= GUI_ALT;
     }
     if (x_modifiers & modifier_info.meta_mask) {
          *modifiers |= GUI_META;
     }
     if (x_modifiers & modifier_info.super_mask) {
          *modifiers |= GUI_SUPER;
     }
     if (x_modifiers & modifier_info.hyper_mask) {
          *modifiers |= GUI_HYPER;
     }
     succeed;
}

def gui_find_hotkey(
     unsigned int x_modifiers,
     GUI_key_code key_code,
     Hotkey** hk)
{
     int i;
     GUI_modifier_mask modifiers;
     
     gui_xmodmask_to_modmask(x_modifiers, &modifiers);
     IFDEBUG(printf("find_window: (mod: %d, key: %d)", modifiers, key_code));
     for (i = 0; i < nhotkeys; i ++) {
          (*hk) = hotkeys[i];
          IFDEBUG(printf("hk: %s, mod: %d, key: %d", (*hk)->hk_name, (*hk)->modifiers, (*hk)->key_code));
          if ((*hk)->modifiers == modifiers && (*hk)->key_code == key_code) {
               succeed;
          }
     }
     *hk = 0;
     fail;
}

defglob gui_run_event_loop(
     int (*exec_hotkey)(Hotkey*))
{
     XEvent an_event;
     yesno done;
     Hotkey* hk;
     
     /* subscribe to KeyPresses (hotkeys!) */
     XSelectInput(gui_display, gui_root, KeyPressMask);
     
     /* perform an events loop */
     done = no;
     while (!done) {
          XNextEvent(gui_display, &an_event);
          if (debug) {
               printf("\n");
          }
          switch (an_event.type) {
          case KeyPress:
               IFDEBUG(printf("X KeyPress event"));
               if (gui_find_hotkey(an_event.xkey.state, an_event.xkey.keycode, &hk)) {
                    exec_hotkey(hk);
               }
               /* exit the application by braking out of the events loop */
               break;

          case KeyRelease:
               IFDEBUG(printf("X KeyRelease event (ignored)"));
               break;
               
          default:
               /* ignore any other event types */
               IFDEBUG(printf("ignored X event (type = %d)", an_event.type));
               break;
          }
     }
     succeed;
}

def gui_store_hotkey(
     Hotkey* hk)
{
     if (hotkeys_size == nhotkeys) {
          Hotkey** new_hotkeys;
          
          new_hotkeys = realloc(hotkeys, sizeof(Hotkey*) * (hotkeys_size * 2 + 1));
          if (new_hotkeys) {
               hotkeys = new_hotkeys;
               hotkeys_size = hotkeys_size * 2 + 1;
          } else {
               fail;
          }
     }
     hotkeys[nhotkeys] = hk;
     nhotkeys ++;
     succeed;
}

defglob gui_register_hotkey(
     Hotkey* hk)
{
     GUI_modifier_mask modifier_mask;
     GUI_key_code key_code;
     unsigned int x_mod_mask;
     
     if (!gui_store_hotkey(hk)) {
          fail;
     }
     modifier_mask = hk->modifiers;
     key_code = hk->key_code;
     gui_modmask_to_xmodmask(modifier_mask, &x_mod_mask);
     gui_clear_error();
     XGrabKey(
          /* display */gui_display
          /* keycode */,key_code/*XKeysymToKeycode(gui_display, XK_a)*/
          /* modifiers */,x_mod_mask/*ControlMask*/
          /* grab_window */,gui_root
          /* owner_events */,False
          /* pointer_mode */,GrabModeAsync
          /* keyboard_mode */,GrabModeAsync);
     XSync(gui_display, False);
     if (gui_error_occured) {
          nhotkeys --;
          hotkeys[nhotkeys] = 0;
          fail;
     } else {
          succeed;
     }
}

defglob gui_unregister_hotkeys()
{
     int i;
     
     XUngrabKey(gui_display, AnyKey, AnyModifier, gui_root);
     for (i = 0; i < nhotkeys; i ++) {
          IFDEBUG(printf("hotkeys_free: %d", i));
          hotkey_free(hotkeys + i);
          hotkeys[i] = 0;
     }
     free(hotkeys);
     hotkeys = 0;
     nhotkeys = 0;
     hotkeys_size = 0;
     succeed;
}

defglob gui_bring_to_front(
     GUI_window w)
{
     IFDEBUG(printf("bring window to front %ld", w));
     XMapWindow(gui_display, w);
     XRaiseWindow(gui_display, w);
     succeed;
}

defglob gui_is_visible(
     GUI_window w)
{
     XWindowAttributes window_attributes;

     if (XGetWindowAttributes(gui_display, w, &window_attributes)) {
          return IsViewable == window_attributes.map_state;
     }
     fail;
}

defglob gui_iconify(
     GUI_window w)
{
     IFDEBUG(printf("iconify window: %ld", w));
     XLowerWindow(gui_display, w);
     XUnmapWindow(gui_display, w);
     succeed;
}

def gui_get_class_hint(
     GUI_window w,
     XClassHint** class_hint)
{
     *class_hint = XAllocClassHint();
     if (0 == *class_hint) {
          fail;
     }
     
     if (!XGetClassHint(gui_display, w, *class_hint)) {
          gui_free_class_hint(class_hint);
          fail;
     }
     
     succeed;
}

def gui_free_class_hint(
     XClassHint** class_hint)
{
     if ((*class_hint)->res_name) {
          XFree((*class_hint)->res_name);
     }
     if ((*class_hint)->res_class) {
          XFree((*class_hint)->res_class);
     }
     XFree(*class_hint);
     (*class_hint) = 0;
     succeed;
}

defglob gui_get_window_class(
     GUI_window w,
     char** class)
{
     XClassHint* class_hint;
     
     *class = 0;
     if (!gui_get_class_hint(w, &class_hint)) {
          fail;
     }
     
     if (class_hint->res_class) {
          *class = strdup(class_hint->res_class);
     }
     
     gui_free_class_hint(&class_hint);
     
     if (*class) {
          succeed;
     } else {
          fail;
     }
}

defglob gui_get_window_name(
     GUI_window w,
     char** name)
{
     XClassHint* class_hint;
     
     *name = 0;
     if (!gui_get_class_hint(w, &class_hint)) {
          fail;
     }
     
     if (class_hint->res_name) {
          *name = strdup(class_hint->res_name);
     }
     
     gui_free_class_hint(&class_hint);
     
     if (*name) {
          succeed;
     } else {
          fail;
     }
}

defglob gui_get_window_title(
     GUI_window w,
     char** title)
{
     char* wmname;
     
     if (XFetchName(gui_display, w, &wmname)) {
          assert(wmname);
          *title = strdup(wmname);
          XFree(wmname);
          succeed;
     } else {
          *title = 0;
          fail;
     }
}

defglob gui_get_window_command(
     GUI_window w,
     char** command)
{
     char **argv;
     int argc;
     
     *command = 0;
     if (XGetCommand(gui_display, w, &argv, &argc)) {
          int i;
          int* lengths = 0;
          int length;
          yesno ok;
          
          lengths = malloc(argc * sizeof(int));
          ok = (lengths != 0);
          if (ok) {
               length = 0;
               IFDEBUG(printf("argc: %d", argc));
               for (i = 0; i < argc; i ++) {
                    IFDEBUG(printf("argv[%d] = %s", i, argv[i]));
                    lengths[i] = strlen(argv[i]);
                    length += lengths[i] + 1;
               }
               IFDEBUG(printf("length: %d", length));
               *command = malloc(length);
               ok = (*command != 0);
          }
          if (ok) {
               length = 0;
               for (i = 0; i < argc; i ++) {
                    if (i > 0) {
                         (*command)[length] = ' ';
                         length ++;
                    }
                    strcpy((*command) + length, argv[i]);
                    length += lengths[i];
                    (*command)[length] = '\0';
                    IFDEBUG(printf("%d: %s", i, *command));
               }
               IFDEBUG(printf("length: %d", strlen(*command)));
          }
          if (lengths) {
               free(lengths);
          }
          if (!ok && *command) {
               free(*command);
               *command = 0;
          }
          XFreeStringList(argv);
          return ok;
     } else {
          fail;
     }
}

defglob gui_get_window_list(
     GUI_window** window_list,
     int* n)
{
     Window root_return;
     Window parent_return;
     Window* children_return;
     unsigned int nchildren_return;
     unsigned int i;
     
     if (
          XQueryTree(
               gui_display,
               gui_root,
               &root_return,
               &parent_return,
               &children_return,
               &nchildren_return)) {
          IFDEBUG(printf("%s: root: %ld, parent: %ld", prog_name, root_return, parent_return));
          (*n) = nchildren_return;
          if ((*n) > 0) {
               *window_list = malloc(sizeof(GUI_window) * nchildren_return);
               assert(*window_list);
          } else {
               *window_list = 0;
          }
          for (i = 0; i < (*n); i ++) {
               (*window_list)[i] = children_return[i];
          }
          if (children_return) {
               XFree(children_return);
          }
          succeed;
     } else {
          fail;
     }
}

def gui_check_modifier_for_known_keysyms(
     XModifierKeymap* mods,
     int row,
     unsigned int mask)
{
     KeyCode* mod_key;
     int col;                   /* the column in the modifier table */
     
     mod_key = mods->modifiermap + row * mods->max_keypermod;
     
     for (col = 0; col < mods->max_keypermod && 0 != mod_key[col]; col ++) {
          switch (XKeycodeToKeysym(gui_display, mod_key[col], 0)) {
          case XK_Meta_L:
          case XK_Meta_R:
               modifier_info.meta_mask |= mask;
               IFDEBUG(printf("Found Meta on %d", modifier_info.meta_mask));
               break;
                    
          case XK_Alt_L:
          case XK_Alt_R:
               modifier_info.alt_mask |= mask;
               IFDEBUG(printf("Found Alt on %d", modifier_info.alt_mask));
               break;
                    
          case XK_Super_L:
          case XK_Super_R:
               modifier_info.super_mask |= mask;
               IFDEBUG(printf("Found Super on %d", modifier_info.super_mask));
               break;
                    
          case XK_Hyper_L:
          case XK_Hyper_R:
               modifier_info.hyper_mask |= mask;
               IFDEBUG(printf("Found Hyper on %d", modifier_info.hyper_mask));
               break;
                    
          default:
               break;
          }
     }
     succeed;
}

/* figure out what keysyms are attached to what modifiers */
def gui_init_modifier_map()
{
     XModifierKeymap* mods;
     
     modifier_info.meta_mask = 0;
     modifier_info.alt_mask = 0;
     modifier_info.super_mask = 0;
     modifier_info.hyper_mask = 0;
     
     mods = XGetModifierMapping(gui_display);
     
     /* gui_check_modifier_for_known_keysyms(mods, 0, ShiftMask); */
     /* gui_check_modifier_for_known_keysyms(mods, 1, LockMask); */
     /* gui_check_modifier_for_known_keysyms(mods, 2, ControlMask); */
     gui_check_modifier_for_known_keysyms(mods, 3, Mod1Mask);
     gui_check_modifier_for_known_keysyms(mods, 4, Mod2Mask);
     gui_check_modifier_for_known_keysyms(mods, 5, Mod3Mask);
     gui_check_modifier_for_known_keysyms(mods, 6, Mod4Mask);
     gui_check_modifier_for_known_keysyms(mods, 7, Mod5Mask);
     
     /* if we couldn't find any meta keys, accept any alt keys as meta keys */
/*       if (!modifier_info.meta_mask) { */
/*            modifier_info.meta_mask = modifier_info.alt_mask; */
/*            modifier_info.alt_mask = 0; */
/*       } */
     
     /* if some keys are both alt and meta, make them just meta, not alt */
     if (modifier_info.alt_mask & modifier_info.meta_mask) {
          modifier_info.alt_mask &= ~modifier_info.meta_mask;
     }
     
     XFreeModifiermap(mods);
     succeed;
}

defglob gui_message(
     const char* title,
     const char* msg)
{
     printf("%s: %s: %s\n", prog_name, title, msg);
     succeed;
}

/* end of file */
